package com.gestalt.sip.utilities.message;

import org.apache.log4j.Logger;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * @author : csturtz
 */
public class DigestClientAuthenticationMethod implements ClientAuthenticationMethod {
    private static final Logger LOG =
                Logger.getLogger(DigestClientAuthenticationMethod.class);

        private String realm;
        private String userName;
        private String uri;
        private String nonce;
        private String password;
        private String method;
        private String cnonce;
        private MessageDigest messageDigest;

        /**
         * to hex converter.
         */
        private static final char[] toHex = {'0', '1', '2', '3', '4', '5', '6',
                '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

        /**
         * convert an array of bytes to an hexadecimal string.
         * @param byteArray bytes array to convert to a hexadecimal string
         * @return hexString string
         */
        public static String toHexString(byte[] byteArray) {
            int pos = 0;
            char[] hexString = new char[byteArray.length * 2];
            for (int i = 0; i < byteArray.length; i++) {
                hexString[pos++] = toHex[(byteArray[i] >> 4) & 0x0F];
                hexString[pos++] = toHex[byteArray[i] & 0x0f];
            }
            return new String(hexString);
        }

        /**
         * Initialize the fields the the response algortihm will use to reply to the
         * authentication challenge.
         * @param realm Realm
         * @param userName Username
         * @param uri URI
         * @param nonce Nonce
         * @param password Password
         * @param method Method
         * @param cnonce CNonce
         * @param algorithm Not used.
         * @throws Exception exception
         */
        public void initialize(String realm, String userName, String uri,
                               String nonce, String password, String method,
                               String cnonce, String algorithm) throws Exception {

            if (realm == null) {
                throw new Exception("The realm parameter is null");
            }
            this.realm = realm;
            if (userName == null) {
                throw new Exception("The userName parameter is null");
            }
            this.userName = userName;
            if (uri == null) {
                throw new Exception("The uri parameter is null");
            }
            this.uri = uri;
            if (nonce == null) {
                throw new Exception("The nonce parameter is null");
            }
            this.nonce = nonce;
            if (password == null) {
                throw new Exception("The password parameter is null");
            }
            this.password = password;
            if (method == null) {
                throw new Exception("The method parameter is null");
            }
            this.method = method;
            this.cnonce = cnonce;

            try {
                messageDigest = MessageDigest.getInstance(algorithm);
            } catch (NoSuchAlgorithmException ex) {
                LOG.debug("DigestClientAuthenticationMethod, initialize(): "
                        + "ERROR: Digest algorithm does not exist.");
                throw new Exception("ERROR: Digest algorithm does not exist.", ex);
            }
        }

        /**
         * Generate the response using the users credentials and the nonce.
         * @return String The Response
         */
        public String generateResponse() {
            if (userName == null) {
                LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                        + "ERROR: no userName parameter");
                return null;
            }
            if (realm == null) {
                LOG.debug("DEBUG, DigestClientAuthenticationMethod, "
                        + "generateResponse(): ERROR: no realm parameter");
                return null;
            }

            LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                    + "Trying to generate a response for the user: "
                    + userName + " , with " + "the realm: " + realm);

            if (password == null) {
                LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                        + "ERROR: no password parameter");
                return null;
            }
            if (method == null) {
                LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                        + "ERROR: no method parameter");
                return null;
            }
            if (uri == null) {
                LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                        + "ERROR: no uri parameter");
                return null;
            }
            if (nonce == null) {
                LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                        + "ERROR: no nonce parameter");
                return null;
            }
            if (messageDigest == null) {
                LOG.debug("DigestClientAuthenticationMethod, generateResponse(): "
                        + "ERROR: the algorithm is not set");
                return null;
            }

            /*******    GENERATE RESPONSE      ************************************/
            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), "
                    + "userName:" + userName + "!");
            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), "
                    + "realm:" + realm + "!");
            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), "
                    + "password:" + password + "!");
            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), "
                    + "uri:" + uri + "!");
            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), "
                    + "nonce:" + nonce + "!");
            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), "
                    + "method:" + method + "!");

            String A1 = userName + ":" + realm + ":" + password;
            String A2 = method.toUpperCase() + ":" + uri;
            byte[] mdbytes = messageDigest.digest(A1.getBytes());
            String HA1 = toHexString(mdbytes);

            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), HA1:"
                    + HA1 + "!");
            mdbytes = messageDigest.digest(A2.getBytes());
            String HA2 = toHexString(mdbytes);

            LOG.debug("DigestClientAuthenticationMethod, generateResponse(), HA2:"
                    + HA2 + "!");
            String KD = HA1 + ":" + nonce;
            if (cnonce != null) {
                KD += ":" + cnonce;
            }
            KD += ":" + HA2;
            mdbytes = messageDigest.digest(KD.getBytes());
            String response = toHexString(mdbytes);

            LOG.debug("DigestClientAlgorithm, generateResponse(): "
                    + "response generated: " + response);

            return response;
        }

}
