package com.gestalt.sip.utilities.message;

import org.apache.log4j.Logger;

import javax.sip.*;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.header.*;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;
import java.util.*;
import java.text.ParseException;

public class RequestGenerator {

    private static final Logger log = Logger.getLogger(RequestGenerator.class);

    private int cseq = 1;

    private MessageFactory messageFactory;
    private HeaderFactory headerFactory;
    private AddressFactory addressFactory;
    private SipProvider sipProvider;
    private String sipURIUser;                  // username
    private String userURI;                     // username@host
    private SipURI userSipURI;
    private String contactIPAddress;
    private String signalingTransport;
    private String outboundProxyIPAddress;
    private int listeningPointPort;
    private int proxyPort;

    private boolean allowToHeaderPort = false;       // default to false

    public RequestGenerator(MessageFactory messageFactory, HeaderFactory headerFactory, AddressFactory addressFactory,
    		SipProvider sipProvider, String sipURIUser, String userURI, String contactIPAddress,
    		String signalingTransport, String outboundProxyIPAddress, int listeningPointPort, int proxyPort) {
    	this.messageFactory = messageFactory;
    	this.headerFactory = headerFactory;
    	this.addressFactory = addressFactory;
    	this.sipProvider = sipProvider;
    	this.sipURIUser = sipURIUser;
    	this.userURI = userURI;
    	this.contactIPAddress = contactIPAddress;
    	this.signalingTransport = signalingTransport;
    	this.outboundProxyIPAddress = outboundProxyIPAddress;
    	this.listeningPointPort = listeningPointPort;
    	this.proxyPort = proxyPort;
        try {
            this.userSipURI = this.addressFactory.createSipURI(this.sipURIUser,this.outboundProxyIPAddress);
        } catch (Exception e) {
            log.error("Exception constructing RequestGenerator!", e);
        }
    }

    private int getCseq() {
        cseq = cseq + 1;
        return cseq;
    }

    public ClientTransaction sendRequest(Request request, Dialog dialog, ClientTransaction clientTransaction, boolean useClientTransaction, boolean useSipProvider) throws Exception {
        log.debug("RequestGenerator.sendRequest()");
        if (useClientTransaction) {
            if (clientTransaction == null) {
                clientTransaction = sipProvider.getNewClientTransaction(request);
            }
            if (log.isDebugEnabled()) {
                log.debug("about to send request:\n" + request);
            }
            if (dialog != null) {
                dialog.sendRequest(clientTransaction);
            }
            else {
                clientTransaction.sendRequest();
            }
            if (log.isDebugEnabled()) {
                log.debug("request sent:\n" + request);
            }
            return clientTransaction;
        }
        if (useSipProvider) {
            if (log.isDebugEnabled()) {
                log.debug("about to send request:\n" + request);
            }
            sipProvider.sendRequest(request);
            if (log.isDebugEnabled()) {
                log.debug("request sent:\n" + request);
            }
        }
        else {
            if (log.isDebugEnabled()) {
                log.debug("about to send request:\n" + request);
            }

//            TransportConduit conduit = ((TransactionImpl) clientTransaction).getConduit();
//            SipStack sipStack = ((TransactionImpl) clientTransaction).getStack();
//            TransportLayerFactory.getLayer(sipStack).sendRequest(conduit, request);

            this.sipProvider.sendRequest(request);

            if (log.isDebugEnabled()) {
                log.debug("request sent:\n" + request);
            }
        }
        return null;
    }

    protected SipURI generateURI(String callee, boolean addPort, boolean addTransport) throws Exception {
        String calleeURI = callee.trim();
        String user = SipUtil.getUserFromURI(calleeURI, 0);
        String host = SipUtil.getHostFromURI(calleeURI, 0);
        SipURI uri = generateURI(user, host, addPort, addTransport);
        if (log.isDebugEnabled()) {
            log.debug("SipURI Generated: " + uri.toString());
        }
        return uri;
    }

    protected SipURI generateURI(String user, String host, boolean addPort, boolean addTransport) throws Exception {
        SipURI requestURI = addressFactory.createSipURI(user, host);
        if (addPort) {
            requestURI.setPort(proxyPort);
        }
        if (addTransport) {
            requestURI.setTransportParam(signalingTransport);
        }
        if (log.isDebugEnabled()) {
            log.debug("SipURI Generated: " + requestURI.toString());
        }
        return requestURI;
    }

    private CallIdHeader generateCallId() {
        return sipProvider.getNewCallId();
    }

    protected CSeqHeader generateCSeqHeader(int sequenceNumber, String requestName) throws Exception {
        return headerFactory.createCSeqHeader(sequenceNumber, requestName);
    }

    protected FromHeader generateFromHeader(Address address, String tag) throws Exception {
        return headerFactory.createFromHeader(address, tag);
    }

    protected FromHeader generateFromHeader(String user, String host, boolean addPort, boolean addTransport, String tag) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.generateFromHeader()");
        }
        SipURI fromURI = generateURI(user, host, addPort, addTransport);
        Address fromAddress = addressFactory.createAddress(fromURI);
        return generateFromHeader(fromAddress, tag);
    }

    protected ToHeader generateToHeader(Address address, String tag) throws Exception {
        return headerFactory.createToHeader(address, tag);
    }

    protected ToHeader generateToHeader(String user, String host, boolean addPort, boolean addTransport, String tag) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.generateToHeader()");
        }
        SipURI toURI = generateURI(user, host, addPort, addTransport);
        if (log.isDebugEnabled()) {
            log.debug("Creating Address from URI: " + toURI);
        }
        Address toAddress = addressFactory.createAddress(toURI);
        return generateToHeader(toAddress, tag);
    }

    protected ViaHeader generateViaHeader(String ipAddress, int port, String transport, boolean generateBranchId) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.generateViaHeader()");
        }
        ViaHeader viaHeader = headerFactory.createViaHeader(ipAddress, port, transport, null);
        if (generateBranchId) {
            viaHeader.setBranch(SipUtil.generateBranchId());
        }
        return viaHeader;
    }

    protected List<ViaHeader> generateViaHeaderList(String ipAddress, int port, String transport, boolean generateBranchId) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.generateViaHeaderList()");
        }
        List<ViaHeader> viaHeaders = new ArrayList<ViaHeader>();
        viaHeaders.add(generateViaHeader(ipAddress, port, transport, generateBranchId));
        return viaHeaders;
    }

    protected MaxForwardsHeader generateMaxForwardsHeader(int maxForwards) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.generateMaxForwardsHeader()");
        }
        return headerFactory.createMaxForwardsHeader(maxForwards);
    }

    protected ContactHeader generateContactHeader(String user, String host, boolean addPort, boolean addTransport) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.generateContactHeader()");
        }
        SipURI contactURI = generateURI(user, host, false, false);
        contactURI.setPort(listeningPointPort);
        Address contactAddress = addressFactory.createAddress(contactURI);
        return headerFactory.createContactHeader(contactAddress);
    }

    public Request createRequest(String requestName, SipURI requestURI, SipURI callerURI, SipURI calleeURI, boolean addProxyRouteHeader) throws Exception {
//        ListeningPoint listeningPoint = sipProvider.getListeningPoint();
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createRequest()");
        }
        CallIdHeader callIdHeader = generateCallId();
        CSeqHeader cSeqHeader = generateCSeqHeader(getCseq(), requestName);
//        FromHeader fromHeader = generateFromHeader(callerURI.getUser(), callerURI.getHost(), true, false, SipUtil.getRandomId());

        // try w/o adding port to FROM header
        FromHeader fromHeader = generateFromHeader(callerURI.getUser(), callerURI.getHost(), false, false, SipUtil.getRandomId());

        // ToHeader
        String calleeHost = calleeURI.getHost();
        if (allowToHeaderPort) {
            if (calleeHost.indexOf(":") == -1 && calleeHost.equals(callerURI.getHost())) {
                calleeHost += ":" + proxyPort;
            }
        }
        ToHeader toHeader = generateToHeader(calleeURI.getUser(), calleeHost, false, false, null);

        List<ViaHeader> viaHeaders = generateViaHeaderList(contactIPAddress, listeningPointPort, signalingTransport, true);
        MaxForwardsHeader maxForwardsHeader = generateMaxForwardsHeader(70);

        Request request = messageFactory.createRequest(requestURI, requestName, callIdHeader, cSeqHeader, fromHeader, toHeader, viaHeaders, maxForwardsHeader);

        ContactHeader contactHeader = generateContactHeader(callerURI.getUser(), contactIPAddress, false, false);
        request.addHeader(contactHeader);

        // Route Header
        if (addProxyRouteHeader) {
            SipURI routeURI = addressFactory.createSipURI(null, outboundProxyIPAddress + ":" + proxyPort);
            routeURI.setLrParam();
            Address routeAddress = addressFactory.createAddress(routeURI);
            RouteHeader routeHeader = headerFactory.createRouteHeader(routeAddress);
            request.addHeader(routeHeader);
        }

        return request;
    }

    public Request createInviteRequest(SipURI caller, String callee, String sdp) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createInviteRequest()");
        }
        SipURI requestURI = generateURI(callee, false, false);
        Request invite = createRequest(Request.INVITE, requestURI, caller, requestURI, false);
        if (sdp != null) {
            ContentTypeHeader contentTypeHeader = headerFactory.createContentTypeHeader("application", "sdp");
            invite.setContent(sdp, contentTypeHeader);
        }
        return invite;
    }

    public Request createSubscribeRequest(SipURI caller, String callee, int expires) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createSubscribeRequest()");
        }
        SipURI requestURI = generateURI(callee, true, true);
        Request subscribe = createRequest(Request.SUBSCRIBE, requestURI, caller, requestURI, false);
        subscribe.addHeader(headerFactory.createExpiresHeader(expires));
        subscribe.addHeader(headerFactory.createEventHeader("presence"));
        subscribe.addHeader(headerFactory.createAcceptHeader("application", "xpidf+xml"));
        subscribe.setContentLength(headerFactory.createContentLengthHeader(0));
        return subscribe;
    }

    public Request createUnsubscribeRequest(SipURI caller, String callee) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createUnsubscribeRequest()");
        }
        SipURI requestURI = generateURI(callee, true, true);
        Request subscribe = createRequest(Request.SUBSCRIBE, requestURI, caller, requestURI, false);
        subscribe.setHeader(headerFactory.createExpiresHeader(0));
        return subscribe;
    }

    public Request createRegisterRequest(SipURI caller, int expires) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createRegisterRequest()");
        }
        SipURI requestURI = generateURI("@" + outboundProxyIPAddress, true, false);
        Request request = createRequest(Request.REGISTER, requestURI, caller, caller, false);

        ContactHeader ch = ((ContactHeader) request.getHeader(ContactHeader.NAME));
        if (ch != null) {
            ch.setExpires(expires);
        }

        request.addHeader(headerFactory.createExpiresHeader(expires));
        request.addHeader(headerFactory.createEventHeader("registration"));
        request.setContentLength(headerFactory.createContentLengthHeader(0));
        return request;
    }

    public Request createRequestWithAuthentication(String username, String password, String proxyhost, Response response, Request newRequest) {
        Header authHeader = createAuthorizationHeader(username, password, proxyhost, response);

        if (authHeader == null) {
            log.debug("IMUserAgent, processResponse(), Proxy-Authorization \" +\n" +
                    " header is null, the request is not resent");
        } else {
            newRequest.setHeader(authHeader);
        }

        return newRequest;
    }

    protected Header createAuthorizationHeader(String username, String password, String proxyhost, Response response) {
        try {
            CSeqHeader cseqHeader = (CSeqHeader) response.getHeader(CSeqHeader.NAME);
            String method = cseqHeader.getMethod();

            String cnonce = null;
            String nonce = null;
            String realm = null;
            String qop = null;
            String uri = "sip:" + proxyhost + ":" + this.proxyPort;

            // Proxy-Authorization header:
            ProxyAuthenticateHeader proxyAuthenticateHeader = (ProxyAuthenticateHeader)
                    response.getHeader(ProxyAuthenticateHeader.NAME);

            WWWAuthenticateHeader wwwAuthenticateHeader = (WWWAuthenticateHeader)
                        response.getHeader(WWWAuthenticateHeader.NAME);

            if (proxyAuthenticateHeader != null) {
                nonce = proxyAuthenticateHeader.getNonce();
                realm = proxyAuthenticateHeader.getRealm();
                if (realm == null) {
                    System.out.println("Realm is NULL!");
                    log.debug("AuthenticationProcess, getProxyAuthorizationHeader()," +
                            " ERROR: the realm is not part of the 407 response!");
                    return null;
                }
                cnonce = proxyAuthenticateHeader.getParameter("cnonce");
                qop = proxyAuthenticateHeader.getParameter("qop");

                DigestClientAuthenticationMethod digest = new DigestClientAuthenticationMethod();
                digest.initialize(realm, username, uri, nonce, password, method, cnonce, "MD5");

                ProxyAuthorizationHeader header = headerFactory.createProxyAuthorizationHeader("Digest");

                header.setParameter("username", username);
                header.setParameter("realm", realm);
                header.setParameter("uri", uri);
                header.setParameter("algorithm", "MD5");
                header.setParameter("opaque", "");
                header.setParameter("nonce", nonce);
                header.setParameter("response", digest.generateResponse());
                if (qop != null) {
                    header.setParameter("qop", qop);
                }

                return header;

            } else if (wwwAuthenticateHeader != null) {
                nonce = wwwAuthenticateHeader.getNonce();
                realm = wwwAuthenticateHeader.getRealm();

                if (realm == null) {
                    log.debug("AuthenticationProcess, getProxyAuthorizationHeader()," +
                            " ERROR: the realm is not part of the 401 response!");
                    return null;
                }
                cnonce = wwwAuthenticateHeader.getParameter("cnonce");
                qop = wwwAuthenticateHeader.getParameter("qop");

                DigestClientAuthenticationMethod digest = new DigestClientAuthenticationMethod();
                digest.initialize(realm, username, uri, nonce, password, method, cnonce, "MD5");

                AuthorizationHeader header = headerFactory.createAuthorizationHeader("Digest");

                header.setParameter("username", username);
                header.setParameter("realm", realm);
                header.setParameter("uri", uri);
                header.setParameter("algorithm", "MD5");
                header.setParameter("opaque", "");
                header.setParameter("nonce", nonce);
                header.setParameter("response", digest.generateResponse());
                if (qop != null) {
                    header.setParameter("qop", qop);
                }

                return header;

            } else {
                log.warn("Unable to find an appropriate Authenticate Header");
                return null;
            }

        } catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }
    }

    public Request createUnregisterRequest(SipURI caller) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createUnregisterRequest()");
        }
        SipURI requestURI = generateURI("@" + outboundProxyIPAddress, true, false);
        Request request = createRequest(Request.REGISTER, requestURI, caller, caller, false);
        request.setHeader(headerFactory.createExpiresHeader(0));
        request.setContentLength(headerFactory.createContentLengthHeader(0));
        return request;
    }

    public Request createMessageRequest(SipURI caller, String callee, String message) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createMessageRequest()");
        }
        SipURI requestURI = generateURI(callee, false, false);
        Request request = createRequest(Request.MESSAGE, requestURI, caller, requestURI, false);
        ContentTypeHeader contentTypeHeader = headerFactory.createContentTypeHeader("text", "plain;charset=UTF-8");
        request.setContent(message, contentTypeHeader);
        return request;
    }

    public Request createByeRequest(Dialog dialog, String caller, String callee) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createByeRequest()");
        }
        boolean originatedHere = (caller.equals(userURI));
        Request request = dialog.createRequest(Request.BYE);

        ContactHeader contactHeader = generateContactHeader(SipUtil.getUserFromURI(getUserURI(),0), contactIPAddress, false, false);
        request.setHeader(contactHeader);

        FromHeader newFromHeader = (FromHeader) request.getHeader(FromHeader.NAME);
        newFromHeader.setTag(dialog.getLocalTag());
        newFromHeader.setAddress(dialog.getLocalParty());

        ToHeader newToHeader = (ToHeader) request.getHeader(ToHeader.NAME);
        newToHeader.setTag(dialog.getRemoteTag());
        newToHeader.setAddress(dialog.getRemoteParty());

        if (originatedHere) {
            request.setRequestURI(generateURI(callee, true, false));
        }
        else {
            request.setRequestURI(generateURI(caller, true, false));
        }

        ViaHeader viaHeader = generateViaHeader(contactIPAddress, listeningPointPort, signalingTransport, true);
        request.addHeader(viaHeader);

        int newcseq = getCseq();
        ((CSeqHeader) request.getHeader(CSeqHeader.NAME)).setSequenceNumber(newcseq);

        return request;
    }

    /**
     * Create a CANCEL Request
     * @return
     * @throws Exception
     */
    public Request createCancelRequest(ClientTransaction clientTransaction) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("RequestGenerator.createByeRequest()");
        }

        Request cancelRequest = clientTransaction.createCancel();
        return cancelRequest;
    }

    // With Jain SIP 1.2, the Stack will send ACKs on behalf of UA for non-2xx responses automatically
    // For 2xx responses, we create an ACK off of the dialog
    public Request create2xxAck(ResponseEvent responseEvent) throws RequestGeneratorException {
        Dialog dialog = responseEvent.getDialog();
        CSeqHeader cseqHeader = (CSeqHeader) responseEvent.getResponse().getHeader(CSeqHeader.NAME);
        long sequenceNumber = cseqHeader.getSeqNumber();

        Request ack = null;
        try {
            ack = dialog.createAck(sequenceNumber);
            ((ToHeader) ack.getHeader(ToHeader.NAME)).setTag(dialog.getRemoteTag());
        } catch (InvalidArgumentException e) {
            throw new RequestGeneratorException("CSeqHeader sequence number from original request is invalid (<=0)",e);
        } catch (SipException e) {
            throw new RequestGeneratorException("CSeqHeader does not relate to an INVITE request, OR\n the initial Method responsible for this Dialog was not an INVITE",e);
        } catch (ParseException e) {
            throw new RequestGeneratorException("Exception setting tag on ToHeader when creating ACK Request",e);
        }

        return ack;
    }

    // With Jain SIP 1.2, the Stack will send ACKs on behalf of UA for non-2xx responses automatically
    public void send2xxAck(Request ack, Dialog dialog) throws RequestGeneratorException {
        try {
            dialog.sendAck(ack);
        } catch (SipException e) {
            throw new RequestGeneratorException("SIPException caught trying to send ACK request",e);
        }
    }


    protected void setHeaderTags(Request request, Dialog dialog) throws Exception {
        FromHeader fromHeader = (FromHeader) request.getHeader(FromHeader.NAME);
        fromHeader.setTag(dialog.getLocalTag());
        request.setHeader(fromHeader);

        ToHeader toHeader = (ToHeader) request.getHeader(ToHeader.NAME);
        toHeader.setTag(dialog.getRemoteTag());
        request.setHeader(toHeader);
    }

    public static List<RecordRouteHeader> getRecordRouteHeaderList(Request request) {
        ListIterator recordRouteHeaderIterator = request.getHeaders(RecordRouteHeader.NAME);
        List<RecordRouteHeader> recordRouteHeaderList = new ArrayList<RecordRouteHeader>();
        while (recordRouteHeaderIterator.hasNext()) {
            recordRouteHeaderList.add((RecordRouteHeader) recordRouteHeaderIterator.next());
        }
        return recordRouteHeaderList;
    }

    public static List<RouteHeader> getRouteHeaderList(Request request) {
        ListIterator routeHeaderIterator = request.getHeaders(RouteHeader.NAME);
        List<RouteHeader> routeHeaderList = new ArrayList<RouteHeader>();
        while (routeHeaderIterator.hasNext()) {
            routeHeaderList.add((RouteHeader) routeHeaderIterator.next());
        }
        return routeHeaderList;
    }


    public static void setRouteHeaderList(Request request, List<RouteHeader> routeHeaderList) {
        for (RouteHeader rHeader : routeHeaderList) {
            request.addHeader(rHeader);
        }
    }

    public SipProvider getSipProvider() {
        return this.sipProvider;
    }

    public SipURI getUserSipURI() {
        return this.userSipURI;
    }

    public String getUserURI() {
        return this.userURI;
    }

    /**
     * see allowToHeaderPort()
     * @return value of this.allowToHeaderPort
     */
    public boolean isAllowToHeaderPort() {
        return this.allowToHeaderPort;
    }

    /**
     * This switch will determine whether or not the URI inside any
     * ToHeader that is generated will include a port number.  Typically,
     * a port is not used, however some products may require it.
     * @param usePort
     */
    public void allowToHeaderPort(boolean usePort) {
        this.allowToHeaderPort = usePort;
    }


    /**
     * Public Inner class 
     */
    public class RequestGeneratorException extends Exception {
        public RequestGeneratorException(String message, Exception e) {
            super(message, e);
        }
    }

    public int getNextCSeq() {
        return getCseq();
    }
}