package com.gestalt.sip.utilities.message;

import java.util.List;
import java.util.ListIterator;

import javax.sip.Dialog;
import javax.sip.ServerTransaction;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.header.CallInfoHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.RecordRouteHeader;
import javax.sip.header.ToHeader;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.Logger;

public class ResponseGenerator {
    
    private static final Logger log = Logger.getLogger(ResponseGenerator.class);  
    
    private MessageFactory messageFactory;
    private HeaderFactory headerFactory;
    private AddressFactory addressFactory;
    
    private String sipURIUser;    
    private String httpBusyMessage;
    private String contactIPAddress;
    private int listeningPointPort;
    
    public ResponseGenerator(MessageFactory messageFactory, HeaderFactory headerFactory, AddressFactory addressFactory, String sipURIUser, String httpBusyMessage, String contactIPAddress, int listeningPointPort) {
    	this.messageFactory = messageFactory;
    	this.headerFactory = headerFactory;
    	this.addressFactory = addressFactory;
    	this.sipURIUser = sipURIUser;
    	this.httpBusyMessage = httpBusyMessage;
    	this.contactIPAddress = contactIPAddress;
    	this.listeningPointPort = listeningPointPort;
    }
    

    public void sendResponse(Response response, ServerTransaction serverTransaction) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("Sending response:\n" + response);
        }
        serverTransaction.sendResponse(response);
        if (log.isDebugEnabled()) {
            log.debug("Response sent:\n" + response);
        }
    }

    public Response generateResponse(int responseType, Request request, boolean generateTag, boolean useRequestTagIfFound, String defaultTag) throws Exception {
        Response response = messageFactory.createResponse(responseType, request);

        // Specify the contact Header
        SipURI contactURI = addressFactory.createSipURI(sipURIUser, contactIPAddress);
        ContactHeader contactHeader = headerFactory.createContactHeader(addressFactory.createAddress(contactURI));
        contactURI.setPort(listeningPointPort);
        response.setHeader(contactHeader);

        if (generateTag) {
            ((ToHeader) response.getHeader(ToHeader.NAME)).setTag(SipUtil.getRandomId());
        }
        else {
            if (useRequestTagIfFound) {
                String tag = ((ToHeader) request.getHeader(ToHeader.NAME)).getTag();
                if (tag == null || tag.equals("")) {
                    if (defaultTag == null) {
                        log.debug("Provided defaultTag is NULL!");
                        defaultTag = SipUtil.getRandomId();
                    }
                    tag = defaultTag;
                }
                ((ToHeader) response.getHeader(ToHeader.NAME)).setTag(tag);
            }
            else {
                if (defaultTag == null) {
                    log.debug("Provided defaultTag is NULL!");
                    defaultTag = SipUtil.getRandomId();
                }
                ((ToHeader) response.getHeader(ToHeader.NAME)).setTag(defaultTag);
            }
        }

        // copy the record routes
        ListIterator iterator = response.getHeaders(RecordRouteHeader.NAME);
        if (iterator != null && iterator.hasNext()) {
            log.debug("RecordRoute Headers Already Added!");
        } else {
            List<RecordRouteHeader> recordRouteHeaders = RequestGenerator.getRecordRouteHeaderList(request);
            if ((recordRouteHeaders != null) && (!recordRouteHeaders.isEmpty())) {
                setRecordRouteHeaders(response,recordRouteHeaders);
            }
        }



        if (log.isDebugEnabled()) {
            log.debug("Response ToHeader: " + response.getHeader(ToHeader.NAME).toString());
        }

        return response;
    }

    public Response generateBusyResponse(Request request) throws Exception {
        Response busyHere = generateResponse(Response.BUSY_HERE, request, false, false, null);
        return busyHere;
    }

    public Response generateRequestTerminatedResponse(Request request) throws Exception {
        return generateResponse(Response.REQUEST_TERMINATED,request,true,false,SipUtil.getRandomId());
    }
    
    public Response generateTryingResponse(Request request) throws Exception {
        return generateResponse(Response.TRYING, request, true, false, null);
    }

    public Response generateRingingResponse(Request invite, Response trying) throws Exception {
        return generateResponse(Response.RINGING, invite, false, false, ((ToHeader) trying.getHeader(ToHeader.NAME)).getTag());
    }

    public Response generateRingingResponse(Request invite) throws Exception {
        return generateResponse(Response.RINGING, invite, false, true, SipUtil.getRandomId());
    }

    public Response generateByeOkResponse(Request bye) throws Exception {
        return generateResponse(Response.OK, bye, false, true, SipUtil.getRandomId() );
    }

    public Response generateCancelOkResponse(Request cancel, Dialog dialog) throws Exception {
        String localTag = dialog.getLocalTag();
        if (localTag == null || localTag.equalsIgnoreCase("")) {
            localTag = SipUtil.getRandomId();
        }
        return generateResponse(Response.OK, cancel, false, false, localTag);
    }

    public Response generateMessageOkResponse(Request message) throws Exception {
        return generateResponse(Response.OK, message, false, true, SipUtil.getRandomId());
    }

    public Response generateMessageNotAcceptableResponse(Request message) throws Exception {
        return generateResponse(Response.NOT_ACCEPTABLE, message, false, false, null);
    }

    public Response generateNotifyOkResponse(Request notify) throws Exception {
        return generateResponse(Response.OK, notify, false, true, SipUtil.getRandomId());
    }

    public Response generateInviteTemporarilyUnavailable(Request invite) throws Exception {
        return generateResponse(Response.TEMPORARILY_UNAVAILABLE, invite,true,false,SipUtil.getRandomId());
    }

    public Response generateDeclineInviteResponse(Request invite, Dialog dialog) throws Exception {
        String localTag = dialog.getLocalTag();
        if (localTag == null || localTag.equalsIgnoreCase("")) {
            localTag = SipUtil.getRandomId();
        }
        return generateResponse(Response.DECLINE,invite,false,false,localTag);
    }

    public Response generateInviteOkResponse(Request request, Dialog dialog, Object sdp) throws Exception {
        if (log.isDebugEnabled()) {
            log.debug("ResponseGenerator.generateInviteOkResponse()");
        }

        String localTag = dialog.getLocalTag();
        if (localTag == null || localTag.equalsIgnoreCase("")) {
            localTag = SipUtil.getRandomId();
        }
        Response ok = generateResponse(Response.OK, request, false, false, localTag);

        ContentTypeHeader contentTypeHeader = (ContentTypeHeader) request.getHeader(ContentTypeHeader.NAME);
        if ((contentTypeHeader != null) && (sdp != null)) {
            ok.setContent(sdp, contentTypeHeader);
        }
        else {
            ok.setHeader(contentTypeHeader);
        }

        if (log.isDebugEnabled()) {
            log.debug("Generated OK Response: \n");
            log.debug(ok);
        }
        return ok;
    }

    public static void setRecordRouteHeaders(Response response,
            List<RecordRouteHeader> recordRouteHeaders) {
        for (RecordRouteHeader rrHeader : recordRouteHeaders) {
            response.addHeader(rrHeader);
        }
    }
}